from .abstractShapeManager import shapeManager
def get_min_latitude_of_geom(geometry):
    """Return the minimum latitude of a geometry which can be Polygon or MultiPolygon."""
    if isinstance(geometry, Polygon):
        min_lat = min(point[1] for point in geometry.exterior.coords)
    elif isinstance(geometry, MultiPolygon):
        min_lat = min(point[1] for polygon in geometry.geoms for point in polygon.exterior.coords)
    else:
        raise TypeError("Unsupported geometry type")
    return min_lat
def get_max_latitude_of_geom(geometry):
    """Return the minimum latitude of a geometry which can be Polygon or MultiPolygon."""
    if isinstance(geometry, Polygon):
        min_lat = max(point[1] for point in geometry.exterior.coords)
    elif isinstance(geometry, MultiPolygon):
        min_lat = max(point[1] for polygon in geometry.geoms for point in polygon.exterior.coords)
    else:
        raise TypeError("Unsupported geometry type")
    return min_lat
# Function to check if a geometry's latitude lies within specified bounds, supporting both Polygon and MultiPolygon
def is_within_bounds(geometry, lower_bound, upper_bound):
    def extract_latitudes(polygon):
        """Extract latitudes from a polygon's exterior coordinates."""
        return [coord[1] for coord in polygon.exterior.coords]

    latitudes = []
    if isinstance(geometry, Polygon):
        latitudes = extract_latitudes(geometry)
    elif isinstance(geometry, MultiPolygon):
        for polygon in geometry.geoms:
            latitudes.extend(extract_latitudes(polygon))
    else:
        raise TypeError("Unsupported geometry type")
    # Check if any latitude falls within the specified bounds
    return any(lower_bound <= lat <= upper_bound for lat in latitudes)
def is_key_found(json_obj,key):
    if key in list(json_obj.keys()):
        return True
def search_for_cardinal_keys(point_object):
    derived_js = {}
    for each_js in [get_designation_js(),get_cardinal_keys()]:
        for key,values in each_js.items():
            for value in values:
                for card_key,value in point_object.keys.items():
                    result = if_eq_or_in(key,value,exact=False,case=False,reverse=True)
                    if result:
                        derived_js[key] = value
                    if key in derived_js:
                        break
                if is_key_found(derived_js,key):
                    break
    return derived_js
def get_city_or_county(point):
    geo_data= []
    for point_itter in ['A','B']:
        for location_type in ['city','county']:
            curr_itter = f"{location_type}{point_itter}"
            value = point.get(curr_itter)
            if value:
                logging.info(f"{curr_itter} {value}")
                return get_closest_designation(location_type),value
    return geo_data
def get_latitude_longitude(point_js):
    point_js['latitude']=False
    if str(point_js.get('direction')).lower() in ['north','south','true']:
        point_js['latitude']=True
    return point_js
def get_derived_geom(point_js):
    designation,value= get_city_or_county(point_js)
    df_a = shape_mgr.get_contents(designation,'shp')
    geom = get_any_geom(df_a,designation,column=get_column_name(designation,file_type='shp'))
    if point_js.get('from') == 'center':
        geom = geom.centroid
    point_js['geom']=geom
    return point_js
def derive_cardinal_vars(point_js):
    point_js=get_derived_geom(point_js)
    point_js = is_center(point_js)
    point_js = get_latitude_longitude(point_js)
    point_js['rangeInclusive'] = str(point_b.get('rangeInclusive')).lower() == 'false'
    return point_js
    



from .file_associations import *
from .transform import *
class shapeManager:
    def __init__(self,directory=None):
        self.shapes_repository = {}
        self.dataDir = directory
        self.directories=self.get_directories()
        self.land_mgr = landManager(self.directories)
    def get_geo_data_dir(self,subDir=None):
        dataDir = self.dataDir
        if subDir:
            dataDir= os.path.join(self.dataDir,subDir)
        return dataDir
    def get_directories(self):
        directories = {'zipcodes':"",'cities':"",'counties':""}
        for designation,value in directories.items():
            directories[designation]=self.get_geo_data_dir(subDir=designation)
        return directories
    def get_contents(self,dir_name,file_type,update=None):
        contents = self.shapes_repository.get(dir_name,{}).get(file_type)
        if contents is None or (update is not None and not isinstance(update,bool)):
            contents = self.land_mgr.get_contents(dir_name=dir_name,file_type=file_type,update=update)
            dir_name = str(dir_name)
            if contents is not None:
                if dir_name not in self.shapes_repository:
                    self.shapes_repository[dir_name] = {}
                if file_type not in self.shapes_repository[dir_name]:
                    self.shapes_repository[dir_name][file_type]=None
                contents = contents.to_crs(epsg=4326)
                self.shapes_repository[dir_name][file_type]=contents
        return self.shapes_repository[dir_name][file_type]
    def get_polygon(self,designation,value):
        designation = get_closest_designation(designation)
        column_name = get_column_name(designation)
        geo_df = shape_mgr.get_contents(designation,'shp')
        return self.land_mgr.get_polygon(geo_df,column_name,value)
    def get_column_name(self,designation,file_type='shp'):
        designation = get_closest_designation(designation)
        column_keys = get_column_js().get(designation)
        geo_df = self.get_contents(designation,file_type)
        column_name = [name for name in geo_df.columns.tolist() if name in column_keys]
        if column_name and isinstance(column_name,list) and len(column_name)>0:
            return column_name[0]
        return column_keys_js[designation][0]
    def get_column_list(self,designation,file_type='shp'):
        column_name = get_column_name(designation,file_type)
        geo_df = self.get_contents(designation,file_type)
        if geo_df is not None:
            return geo_df[column_name].tolist()



from shapely.geometry import Point, Polygon, MultiPolygon
def get_position_of_geom(geom, r=max, latitude=True):
    """
    Find the position (either min or max) of a geometry by latitude or longitude.

    Args:
        geom: A shapely geometry object (Polygon, MultiPolygon, or Point).
        r: A function to apply (min or max).
        latitude: If True, find the point by latitude; otherwise, by longitude.

    Returns:
        A tuple representing the coordinates of the desired position.
    """
    def coord_extractor(coord):
        return coord[1] if latitude else coord[0]

    if isinstance(geom, Point):
        return (geom.x, geom.y)
    elif isinstance(geom, Polygon):
        coords = list(geom.exterior.coords)
    elif isinstance(geom, MultiPolygon):
        coords = [coord for polygon in geom.geoms for coord in polygon.exterior.coords]
    else:
        raise ValueError("Unsupported geometry type.")

    # Extract the desired coordinate based on latitude or longitude
    desired_coord = r(coords, key=coord_extractor)
    return desired_coord

def get_extremity(geom_a, geom_b, latitude=True):
    """
    Determine which of two geometries has the most extreme point (north/south or east/west).

    Args:
        geom_a: A shapely geometry object.
        geom_b: Another shapely geometry object.
        latitude: If True, compares by latitude (north/south); otherwise, by longitude (east/west).

    Returns:
        A dictionary containing the geometries and their extreme points.
    """
    comparison_key = 1 if latitude else 0

    # Choose comparison function based on direction
    r = max if latitude else min

    max_a = get_position_of_geom(geom_a, r=r, latitude=latitude)
    max_b = get_position_of_geom(geom_b, r=r, latitude=latitude)

    is_a_more_extreme = max_a[comparison_key] > max_b[comparison_key] if latitude else max_a[comparison_key] < max_b[comparison_key]
    return {
        "max": {
            "geom": 'a' if is_a_more_extreme else 'b',
            "geometry": geom_a if is_a_more_extreme else geom_b,
            "coordinates": max_a if is_a_more_extreme else max_b,
            'direction': latitude
        },
        "min": {
            "geom": 'b' if is_a_more_extreme else 'a',
            "geometry": geom_b if is_a_more_extreme else geom_a,
            "coordinates": max_b if is_a_more_extreme else max_a,
            'direction': latitude
        }
    }

def get_auto_most(geom_a, geom_b, latitude='north'):
    """
    Determine the most extreme geometry based on cardinal direction.

    Args:
        geom_a: A shapely geometry object.
        geom_b: Another shapely geometry object.
        latitude: One of 'north', 'south', 'east', 'west'.

    Returns:
        A dictionary containing the geometries and their extreme points.
    """
    latitude_lower = str(latitude).lower()
    if latitude_lower in ['south', 'west']:
        r = min
    else:
        r = max

    latitude = latitude_lower in ['north', 'south']
    return get_extremity(geom_a, geom_b, latitude=latitude)

def get_range(geom_a, geom_b, latitude='north', rangeInclusiveA=False, rangeInclusiveB=False):
    """
    Determine the range between two geometries based on the specified direction.

    Args:
        geom_a: A shapely geometry object.
        geom_b: Another shapely geometry object.
        latitude: One of 'north', 'south', 'east', 'west'.
        rangeInclusiveA: Boolean specifying whether the first geometry should include its extremity.
        rangeInclusiveB: Boolean specifying whether the second geometry should include its extremity.

    Returns:
        A dictionary containing the directional range limits and their corresponding coordinates.
    """
    reference_js = {"a": {"geom": geom_a, "inclusive": rangeInclusiveA}, "b": {"geom": geom_b, "inclusive": rangeInclusiveB}}
    auto_js = get_auto_most(geom_a, geom_b, latitude=latitude)

    # Determine the reference functions (min or max) based on inclusivity
    r_a = max if reference_js[auto_js['max']['geom']]["inclusive"] else min
    r_b = min if reference_js[auto_js['min']['geom']]["inclusive"] else max

    return {
        "latitude": [auto_js['min']['direction'], auto_js['max']['direction']],
        "range": [
            get_position_of_geom(reference_js[auto_js['min']['geom']]["geom"], r=r_b, latitude=auto_js['min']['direction']),
            get_position_of_geom(reference_js[auto_js['max']['geom']]["geom"], r=r_a, latitude=auto_js['max']['direction'])
        ]
    }

# Function to check if a geometry's latitude lies within specified bounds, supporting both Polygon and MultiPolygon
def is_within_bounds(geometry, lower_bound, upper_bound):
    def extract_latitudes(polygon):
        """Extract latitudes from a polygon's exterior coordinates."""
        return [coord[1] for coord in polygon.exterior.coords]

    latitudes = []
    if isinstance(geometry, Polygon):
        latitudes = extract_latitudes(geometry)
    elif isinstance(geometry, MultiPolygon):
        for polygon in geometry.geoms:
            latitudes.extend(extract_latitudes(polygon))
    else:
        raise TypeError("Unsupported geometry type")

    # Check if any latitude falls within the specified bounds
    return any(lower_bound <= lat <= upper_bound for lat in latitudes)



from osgeo import ogr, osr
from shapely import wkb,Polygon,MultiPolygon
def get_obj_crs(epsg=None,proj_string=None,obj=osr):
    obj_crs = obj.SpatialReference()
    if epsg:
        obj_crs.ImportFromEPSG(epsg)
    else:
        obj_crs.ImportFromWkt(str(proj_string))
    return obj_crs
def setup_transformation(source_epsg=None, target_epsg=None,source_proj_string=None,target_proj_string=None):
    source_crs = get_obj_crs(epsg=source_epsg,proj_string=source_proj_string)
    target_crs = get_obj_crs(epsg=target_epsg,proj_string=target_proj_string)
    return osr.CoordinateTransformation(source_crs,target_crs)
def transform_ogr_geometry(ogr_geom, transform):
    """
    Transform a single ogr.Geometry and return it as a Shapely geometry.
    """
    # Clone the original geometry to avoid modifying it directly
    transformed_ogr_geom = ogr_geom.Clone()
    # Apply the transformation
    transformed_ogr_geom.Transform(transform)
    # Export to WKB
    wkb_data = transformed_ogr_geom.ExportToWkb()
    wkb_data=bytes(wkb_data)
    # Check if wkb_data is a bytes object
    if not isinstance(wkb_data, bytes):
        raise TypeError(f"Expected WKB data to be bytes, got {type(wkb_data).__name__}")

    # Convert the WKB data to a Shapely geometry
    new_shapely_geom = wkb.loads(wkb_data)
    return new_shapely_geom

def transform_geometry(geom, transform):
    """
    Apply a coordinate transformation to any Shapely geometry and return a new geometry.
    """
    # Convert Shapely geometry to ogr.Geometry if necessary
    if isinstance(geom, (Polygon, MultiPolygon)):
        ogr_geom = ogr.CreateGeometryFromWkb(geom.wkb)
    else:
        raise TypeError("Unsupported geometry type")
    
    if isinstance(geom, MultiPolygon):
        # Handle MultiPolygon by transforming each constituent polygon
        transformed_polygons = [transform_ogr_geometry(ogr.CreateGeometryFromWkb(poly.wkb), transform) for poly in geom.geoms]
        return MultiPolygon(transformed_polygons)
    else:
        # Handle single Polygon
        return transform_ogr_geometry(ogr_geom, transform)
def update_geom(gdf,index,transform):
    existing_geom = gdf.at[index, 'geometry']
    return transform_geometry(existing_geom, transform)
    
def update_geometry_in_gdf(gdf, index=None, source_epsg=None, target_epsg=None,source_proj_string=None,target_proj_string=None):
    """
    Update the geometry in a GeoDataFrame at a specific index if the transformed geometry is different.
    """
    # Create transformation
    transform = setup_transformation(source_epsg=source_epsg, target_epsg=target_epsg,source_proj_string=source_proj_string,target_proj_string=target_proj_string)
    # Get the existing geometry and transform it

    if index== None:
        for index in range(len(gdf)):
            new_geom = update_geom(gdf,index,transform)
            if new_geom:
                gdf.at[index, 'geometry'] = new_geom
        return gdf
    else:
        # Check if the new geometry is different from the existing geometry
        if not existing_geom.equals(new_geom):
            gdf.at[index, 'geometry'] = new_geom
            print(f"Updated row {index} with new geometry.")
        else:
            print(f"No update needed for row {index}.")
        return gdf
def update_geop_json(desig,epsg=4326):
    gdj = get_shp(desig).to_crs(epsg=epsg)
    return gdj



from abstract_pandas import *
def clean_list(list_obj):
    while '' in list_obj:
        list_obj.remove('')
    return list_obj
def get_cardinal_keys():
    return {'direction'["north","south","easty","west"],'direction','from':['center',"boundary"],"rangeInclusive":["inclusive","range"]}
def get_column_js():
    return {'zipcodes':['ZIP_CODE'],'cities':['CITY'],'counties':["countyName","COUNTY_NAM","NAME"]}
def get_designation_js():
    return {'cities':['city','cities','citie'],'counties':['counties','countie','county','countys','counties'],'zipcodes':['zips','zipcodes','zip-codes','zip_codes','zip codes','zipcode','zip','zip-code','zip_code','zip code']}
def if_eq_or_in(string,compString,exact=False,case=False,reverse=True):
    if not case:
        string=str(string).lower()
        compString = str(compString).lower()
    if string == compString:
        return True
    if not exact:
        if string in compString:
            return True
    if reverse:
        if compString in string:
            return True
    return False
def is_str_in_list_versa(list_obj,string):
    string_lower = str(string).lower()
    list_lower = [str(obj).lower() for obj in list_obj]
    for obj in lost_lower:
        if obj in string_lower or string_lower in obj:
            return True
    return False
def get_closest_designation(designation):
    if isinstance(designation,gpd.GeoDataFrame):
        columns = get_df(designation,nrows=0)
        for designation_type,values in get_designation_js().items():
            for column in columns:
                for value in values:
                    response = if_eq_or_in(column,value,exact=True,case=False,reverse=False)
                    if response:
                        return designation_type
    designation_lower = str(designation).lower()
    designation_js = get_designation_js()
    for key,values in designation_js.items():
        values.append(key.lower())
        if designation_lower in values:
            return key
    for key,values in designation_js.items():
        values.append(key.lower())
        if is_str_in_list_versa(values,designation_lower):
            return key
def get_any_row(df,value,column=None):
    value = str(value).lower()
    headers =  get_df(df,nrows=0)
    for header in headers:
        values = [val for val in enumerate(df[header].tolist()) if if_eq_or_in(value,val[1])]
        if values:
            return df.iloc[values[0][0]]
def get_any_geom(df,value,column=None):
    value = str(value).lower()
    headers =  get_df(df,nrows=0)
    for header in headers:
        values = [val for val in enumerate(df[header].tolist()) if  if_eq_or_in(value,val[1])]
        if values:
            return df.at[values[0][0],'geometry']
def get_any_header(df,value,column=None):
    value = str(value).lower()
    headers =  get_df(df,nrows=0)
    for header in headers:
        values = [val for val in enumerate(df[header].tolist()) if if_eq_or_in(value,val[1])]
        if values:
            return header
def get_city_or_county(point):
    geo_data= []
    for point_itter in ['A','B']:
        for location_type in ['city','county']:
            curr_itter = f"{location_type}{point_itter}"
            value = point.get(curr_itter)
            if value:
                logging.info(f"{curr_itter} {value}")
                return get_closest_designation(location_type),value
    return geo_data



def get_column_js():
    return {'zipcodes':['ZIP_CODE'],'cities':['CITY'],'counties':["countyName","COUNTY_NAM","NAME"]}
def get_designation_js():
    return {'cities':['city','cities','citie'],'counties':['counties','countie','county','countys','counties'],'zipcodes':['zips','zipcodes','zip-codes','zip_codes','zip codes','zipcode','zip','zip-code','zip_code','zip code']}
def get_column_name(designation,file_type='shp'):
    designation = get_closest_designation(designation)
    column_keys = get_column_js().get(designation)
    geo_df = shape_mgr.get_contents(designation,file_type)
    column_name = [name for name in geo_df.columns.tolist() if name in column_keys]
    if column_name and isinstance(column_name,list) and len(column_name)>0:
        return column_name[0]
    return column_keys_js[designation][0]
def get_column_list(designation,file_type='shp'):
    column_name = get_column_name(designation,file_type)
    geo_df = shape_mgr.get_contents(designation,file_type)
    if geo_df is not None:
        return geo_df[column_name].tolist()
def if_eq_or_in(string,compString,exact=False,case=False,reverse=True):
    if not case:
        string=str(string).lower()
        compString = str(compString).lower()
    if string == compString:
        return True
    if not exact:
        if string in compString:
            return True
    if reverse:
        if compString in string:
            return True
    return False
def get_any_row(df,value,column=None):
    value = str(value).lower()
    headers =  get_df(df,nrows=0)
    for header in headers:
        values = [val for val in enumerate(df[header].tolist()) if if_eq_or_in(value,val[1])]
        if values:
            return df.iloc[values[0][0]]
def get_any_geom(df,value,column=None):
    value = str(value).lower()
    headers =  get_df(df,nrows=0)
    for header in headers:
        values = [val for val in enumerate(df[header].tolist()) if  if_eq_or_in(value,val[1])]
        if values:
            return df.at[values[0][0],'geometry']
def get_any_header(df,value,column=None):
    value = str(value).lower()
    headers =  get_df(df,nrows=0)
    for header in headers:
        values = [val for val in enumerate(df[header].tolist()) if if_eq_or_in(value,val[1])]
        if values:
            return header
